import { createElement, Fragment, useReducer, useRef, useCallback, useEffect, useMemo, useDebugValue, createContext, useContext, Children, isValidElement, cloneElement } from 'react';
import isEqual from 'fast-deep-equal';
import breakLines from 'break-styled-lines';

// Is the windup of type PlayedWindup?
function isPlayedWindup(windup) {
  // If it's not an array it can't be a windup
  if (!Array.isArray(windup)) {
    return false;
  } // If it has less or more than three members it's not a windup


  if (windup.length !== 2) {
    return false;
  } // If its first member is not  anarrays it's not a windup


  if (!Array.isArray(windup[0])) {
    return false;
  }

  return true;
} // Is the member a windup, or just a member type?

function memberIsWindup(member) {
  // If it's not an array it can't be a windup
  if (!Array.isArray(member)) {
    return false;
  } // If it has less or more than three members it's not a windup


  if (member.length !== 3) {
    return false;
  } // If its first or second members are not arrays it's not a windup


  if (!Array.isArray(member[0]) || !Array.isArray(member[1])) {
    return false;
  } // Past here we just have to hope ElementType isn't a windup.


  return true;
} // Is the member a windup, or just a member type?

function playedMemberIsWindup(member) {
  // If it's not an array it can't be a windup
  if (!Array.isArray(member)) {
    return false;
  } // If it has less or more than three members it's not a windup


  if (member.length !== 2) {
    return false;
  } // If its first member is not an array it's not a windup


  if (!Array.isArray(member[0])) {
    return false;
  } // Past here we just have to hope ElementType isn't a windup.


  return true;
} // Create a windup from a given string
// "Hello", metadata -> [[], ["H", "e", "l", "l", "o"], metadata]

function windupFromString(str, metadata) {
  return [[], str.split(""), metadata];
}
function newWindup(arg, metadata) {
  return [[], arg, metadata];
} // Checks if the windup is unplayed
// INCLUDING any nested windups in the remaining part of the windup

function isUnplayed(windup) {
  if (isPlayedWindup(windup)) {
    return false;
  }

  var played = windup[0],
      remaining = windup[1];

  if (played.length > 0) {
    return false;
  }

  return remaining.reduce(function (unplayed, member) {
    if (memberIsWindup(member)) if (memberIsWindup(windup) && unplayed) {
      return isUnplayed(member);
    }
    return unplayed;
  }, true);
} // Checks if a windup has any remaining parts

function isFinished(windup) {
  if (isPlayedWindup(windup)) {
    return true;
  }

  var remaining = windup[1];
  return remaining.length === 0;
} // Turn a Windup into a PlayedWindup

function fastForward(windup) {
  if (isPlayedWindup(windup)) {
    return windup;
  }

  var forwardedWindup = next(windup);

  if (isPlayedWindup(forwardedWindup)) {
    return forwardedWindup;
  }

  return fastForward(forwardedWindup);
} // Turn an in-progress Windup or PlayedWindup into an unplayed Windup

function rewind(windup) {
  if (!isPlayedWindup(windup) && isUnplayed(windup)) {
    return windup;
  }

  var played = windup[0];
  var metadata = isPlayedWindup(windup) ? windup[1] : windup[2];

  var mapRewind = function mapRewind(member) {
    if (memberIsWindup(member)) {
      return rewind(member);
    }

    return member;
  };

  var playedMapRewind = function playedMapRewind(member) {
    if (playedMemberIsWindup(member)) {
      return rewind(member);
    }

    return member;
  };

  if (isPlayedWindup(windup)) {
    return [[], played.map(playedMapRewind), metadata];
  }

  var remaining = windup[1];
  return [[], [].concat(played.map(playedMapRewind), remaining.map(mapRewind)), metadata];
}
function lastPlayedMember(windup) {
  if (isPlayedWindup(windup)) {
    var _played = windup[0];
    var _last = _played[_played.length - 1];
    return _last;
  }

  var played = windup[0],
      remaining = windup[1];
  var playedFromRemaining = remaining.reduce(function (playedEl, member) {
    if (memberIsWindup(member)) {
      if (!isUnplayed(member)) {
        return lastPlayedElement(member);
      }
    }

    return playedEl;
  }, undefined);

  if (playedFromRemaining) {
    return playedFromRemaining;
  }

  var last = played[played.length - 1];
  return last;
}
function lastPlayedElement(windup) {
  var lastPlayed = lastPlayedMember(windup);

  if (playedMemberIsWindup(lastPlayed)) {
    return lastPlayedElement(lastPlayed);
  }

  return lastPlayed;
} // Return the played elements of a windup
// INCLUDING any played elements from nested windups in the remaining part of the windup

function playedElements(windup) {
  if (isPlayedWindup(windup)) {
    var _played2 = windup[0];
    return _played2;
  }

  var played = windup[0],
      remaining = windup[1];
  var firstRemaining = remaining[0];

  if (memberIsWindup(firstRemaining) && !isUnplayed(firstRemaining)) {
    var firstRemainingMetadata = firstRemaining[2];
    return [].concat(played, [[playedElements(firstRemaining), firstRemainingMetadata]]);
  }

  return played;
}
function nextElement(windup) {
  if (isPlayedWindup(windup)) {
    return undefined;
  }

  var remaining = windup[1];
  var nextVal = remaining[0];

  if (memberIsWindup(nextVal)) {
    return nextElement(nextVal);
  }

  return nextVal;
}
function next(windup) {
  // start
  // [[], [[[], ["h", "i"]], [[], ["n", "o"]]]]
  // next
  // [[], [[["h"], ["i"]], ["", ["n", "o"]]]]
  // next
  // [["h", "i"], [["n", "o"]]]
  // next
  // [["h", "i"] ["n", "o"]], []]
  if (isPlayedWindup(windup)) {
    return windup;
  }

  var played = windup[0],
      remaining = windup[1],
      metadata = windup[2];
  var firstRemaining = remaining[0],
      restRemaining = remaining.slice(1);

  if (!firstRemaining) {
    return [played, metadata];
  }

  if (memberIsWindup(firstRemaining)) {
    var nextFirstRemaining = next(firstRemaining);

    if (isPlayedWindup(nextFirstRemaining)) {
      if (restRemaining.length === 0) {
        return [[].concat(played, [nextFirstRemaining]), metadata];
      }

      return [[].concat(played, [nextFirstRemaining]), restRemaining, metadata];
    }

    return [played, [nextFirstRemaining].concat(restRemaining), metadata];
  }

  if (restRemaining.length === 0) {
    return [[].concat(played, [firstRemaining]), metadata];
  }

  return [[].concat(played, [firstRemaining]), restRemaining, metadata];
}

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

function isSuffixedWindup(windup) {
  // If it's not an array it can't be a windup
  if (!Array.isArray(windup)) {
    return false;
  } // If it has less or more than three members it's not a windup


  if (windup.length !== 4) {
    return false;
  } // If its first or second members are not arrays it's not a windup


  if (!Array.isArray(windup[0]) || !Array.isArray(windup[1])) {
    return false;
  } // Past here we just have to hope ElementType isn't a windup.


  return true;
}
function isPlayedSuffixedWindup(windup) {
  // If it's not an array it can't be a windup
  if (!Array.isArray(windup)) {
    return false;
  } // If it has less or more than three members it's not a windup


  if (windup.length !== 3) {
    return false;
  } // If its first or second members are not arrays it's not a windup


  if (!Array.isArray(windup[0]) || !Array.isArray(windup[1])) {
    return false;
  } // Past here we just have to hope ElementType isn't a windup.


  return true;
}
function isUnplayed$1(windup) {
  var played = windup[0],
      remaining = windup[1];

  if (isPlayedSuffixedWindup(windup)) {
    return false;
  }

  if (played.length > 0) {
    return false;
  }

  return remaining.reduce(function (unplayed, member) {
    if (isSuffixedWindup(member)) if (isSuffixedWindup(windup) && unplayed) {
      return isUnplayed$1(member);
    }
    return unplayed;
  }, true);
}
function playedSize(suffixedWindup) {
  var suffixedPlayed = suffixedWindup[0],
      remaining = suffixedWindup[1];
  var sizeOfPlayed = suffixedPlayed.reduce(function (acc, member) {
    if (isPlayedSuffixedWindup(member)) {
      return acc + playedSize(member);
    }

    return acc + 1;
  }, 0);
  var firstRemaining = remaining[0];

  if (firstRemaining && isSuffixedWindup(firstRemaining)) {
    var sizeFromFirstRemaining = playedSize(firstRemaining);
    return sizeOfPlayed + sizeFromFirstRemaining;
  }

  return sizeOfPlayed;
}
function playedElementsWithSuffix(windup) {
  if (isPlayedSuffixedWindup(windup)) {
    var played = windup[0],
        _suffix = windup[1];
    return [].concat(played, _suffix.map(function (member) {
      if (playedMemberIsWindup(member)) {
        return playedWindupToSuffixed(member);
      }

      return member;
    }));
  }

  var suffixedPlayed = windup[0],
      remaining = windup[1],
      suffix = windup[2];
  var middle = remaining.reduce(function (acc, member) {
    if (isSuffixedWindup(member)) {
      var metadata = member[3];
      var asWindup = [[], playedElementsWithSuffix(member), metadata];
      return [].concat(acc, [asWindup]);
    }

    return acc;
  }, []);
  return [].concat(suffixedPlayed, middle, suffix.map(function (member) {
    if (playedMemberIsWindup(member)) {
      return playedWindupToSuffixed(member);
    }

    return member;
  }));
}
function prev(suffixedWindup) {
  if (isPlayedSuffixedWindup(suffixedWindup)) {
    var _played = suffixedWindup[0],
        _suffix2 = suffixedWindup[1],
        _metadata = suffixedWindup[2];

    var _restPlayed = _played.slice(0, -1);

    var _lastPlayed = _played[_played.length - 1];

    if (!_lastPlayed) {
      return [_played, [], _suffix2, _metadata];
    }

    if (isPlayedSuffixedWindup(_lastPlayed)) {
      return [_restPlayed, [prev(_lastPlayed)], _suffix2, _metadata];
    }

    return [_restPlayed, [_lastPlayed], _suffix2, _metadata];
  }

  var played = suffixedWindup[0],
      remaining = suffixedWindup[1],
      suffix = suffixedWindup[2],
      metadata = suffixedWindup[3];
  var firstRemaining = remaining[0],
      restRemaining = remaining.slice(1);

  if (!firstRemaining) {
    var _restPlayed2 = played.slice(0, -1);

    var _lastPlayed2 = played[played.length - 1];

    if (!_lastPlayed2) {
      return [played, [], suffix, metadata];
    }

    if (isPlayedSuffixedWindup(_lastPlayed2)) {
      return [_restPlayed2, [prev(_lastPlayed2)], suffix, metadata];
    }

    return [_restPlayed2, [_lastPlayed2], suffix, metadata];
  } // check if first remaining is an inprogress windup


  if (isSuffixedWindup(firstRemaining) && !isUnplayed$1(firstRemaining)) {
    return [played, [prev(firstRemaining)].concat(restRemaining), suffix, metadata];
  } // if it's not (element or unplayed)
  // take the last element of the played value


  var restPlayed = played.slice(0, -1);
  var lastPlayed = played[played.length - 1];

  if (isPlayedSuffixedWindup(lastPlayed)) {
    return [restPlayed, [prev(lastPlayed)].concat(remaining), suffix, metadata];
  } // if it's an element, just take it


  return [restPlayed, [lastPlayed].concat(remaining), suffix, metadata];
}
function next$1(suffixedWindup) {
  if (isPlayedSuffixedWindup(suffixedWindup)) {
    return suffixedWindup;
  }

  var played = suffixedWindup[0],
      remaining = suffixedWindup[1],
      suffix = suffixedWindup[2],
      metadata = suffixedWindup[3];
  var firstRemaining = remaining[0],
      restRemaining = remaining.slice(1);

  if (!firstRemaining) {
    return [played, suffix, metadata];
  }

  if (isSuffixedWindup(firstRemaining)) {
    var nextFirstRemaining = next$1(firstRemaining);

    if (isPlayedSuffixedWindup(nextFirstRemaining)) {
      if (restRemaining.length === 0) {
        return [[].concat(played, [nextFirstRemaining]), suffix, metadata];
      }

      return [[].concat(played, [nextFirstRemaining]), restRemaining, suffix, metadata];
    }

    return [played, [nextFirstRemaining].concat(restRemaining), suffix, metadata];
  }

  if (restRemaining.length === 0) {
    return [[].concat(played, [firstRemaining]), suffix, metadata];
  }

  return [[].concat(played, [firstRemaining]), restRemaining, suffix, metadata];
}
function fastForward$1(windup) {
  var forwardedWindup = next$1(windup);

  if (isPlayedSuffixedWindup(forwardedWindup)) {
    return forwardedWindup;
  }

  return fastForward$1(forwardedWindup);
}
function windupToSuffixed(windup) {
  var played = windup[0],
      remainder = windup[1],
      metadata = windup[2];
  return [played.map(function (member) {
    if (!playedMemberIsWindup(member)) {
      return member;
    }

    return playedWindupToSuffixed(member);
  }), remainder.map(function (member) {
    if (!memberIsWindup(member)) {
      return member;
    }

    return windupToSuffixed(member);
  }), [], metadata];
}
function playedWindupToSuffixed(windup) {
  var played = windup[0],
      metadata = windup[1];
  return [played.map(function (member) {
    if (!playedMemberIsWindup(member)) {
      return member;
    }

    return playedWindupToSuffixed(member);
  }), [], metadata];
}
var MemberKind;

(function (MemberKind) {
  MemberKind[MemberKind["Played"] = 0] = "Played";
  MemberKind[MemberKind["InProgress"] = 1] = "InProgress";
})(MemberKind || (MemberKind = {}));

function allMembers(windup) {
  var played = isPlayedWindup(windup) ? windup[0] : windup[0];
  var remainder = !isPlayedWindup(windup) ? windup[1] : [];
  return [].concat(played.map(function (member) {
    return playedMemberIsWindup(member) ? {
      member: member,
      kind: MemberKind.Played,
      isWindup: true
    } : {
      member: member,
      kind: MemberKind.Played,
      isWindup: false
    };
  }), remainder.map(function (member) {
    return memberIsWindup(member) ? {
      member: member,
      kind: MemberKind.InProgress,
      isWindup: true
    } : {
      member: member,
      kind: MemberKind.InProgress,
      isWindup: false
    };
  }));
} // Check if members are equivalent

function isMemberEquivalent(memberA, memberB) {
  if (!memberA.isWindup && memberB.isWindup || memberA.isWindup && !memberB.isWindup) {
    return false;
  }

  if (memberA.isWindup && memberB.isWindup) {
    // TODO: look at metadata here
    var aMetadata = getMetadata(memberA.member);
    var bMetadata = getMetadata(memberB.member);

    if (!isEqual(aMetadata, bMetadata)) {
      return false;
    }

    var aMembers = allMembers(memberA.member);
    var bMembers = allMembers(memberB.member);

    if (aMembers.length !== bMembers.length) {
      return false;
    }

    var zipped = zipMembersFromLeft(aMembers, bMembers);
    return zipped.every(function (_ref) {
      var a = _ref[0],
          b = _ref[1];
      return b ? isMemberEquivalent(a, b) : false;
    });
  }

  return isEqual(memberA.member, memberB.member);
}

function matchesFromLeft(zipped) {
  return zipped.reduce(function (matches, _ref2) {
    var aMember = _ref2[0],
        bMember = _ref2[1];

    if (bMember && isMemberEquivalent(aMember, bMember) && matches.unmatched.length === 0) {
      return _extends({}, matches, {
        matched: [].concat(matches.matched, [aMember])
      });
    }

    return _extends({}, matches, {
      unmatched: [].concat(matches.unmatched, [[aMember, bMember]])
    });
  }, {
    matched: [],
    unmatched: []
  });
}

function matchesFromRight(zipped) {
  return zipped.reduce(function (matches, _ref3) {
    var aMember = _ref3[0],
        bMember = _ref3[1];

    if (bMember && isMemberEquivalent(aMember, bMember) && matches.unmatched.length === 0) {
      return _extends({}, matches, {
        matched: [aMember].concat(matches.matched)
      });
    }

    return _extends({}, matches, {
      unmatched: [[aMember, bMember]].concat(matches.unmatched)
    });
  }, {
    matched: [],
    unmatched: []
  });
}

function zipMembersFromLeft(baseMembers, againstMembers) {
  return baseMembers.map(function (member, i) {
    return [member, againstMembers[i]];
  });
}

function zipMembersFromRight(baseMembers, againstMembers) {
  var zippedFromEnd = [];

  for (var i = 0; i <= baseMembers.length - 1; i++) {
    zippedFromEnd.push([baseMembers[baseMembers.length - 1 - i], againstMembers[againstMembers.length - 1 - i]]);
  }

  return zippedFromEnd;
}

function prefixFromPair(pair) {
  var baseMember = pair[0],
      againstMember = pair[1];

  if (!baseMember.isWindup) {
    return undefined;
  }

  if (!(againstMember != null && againstMember.isWindup)) {
    return undefined;
  }

  var metadata = getMetadata(baseMember.member);
  var baseMembers = allMembers(baseMember.member);
  var againstMembers = allMembers(againstMember.member);
  var zipped = zipMembersFromLeft(baseMembers, againstMembers);

  var _matchesFromLeft = matchesFromLeft(zipped),
      matched = _matchesFromLeft.matched,
      unmatchedFromLeft = _matchesFromLeft.unmatched;

  var played = matched.map(function (member) {
    if (member.kind === MemberKind.InProgress && member.isWindup) {
      return playedWindupToSuffixed(fastForward(member.member));
    }

    if (member.kind === MemberKind.Played && member.isWindup) {
      return playedWindupToSuffixed(member.member);
    }

    return member.member;
  });
  var firstUnmatchedFromLeft = unmatchedFromLeft[0];
  var maybePrefix = prefixFromPair(firstUnmatchedFromLeft);
  var fromLeftBaseMembers = unmatchedFromLeft.map(function (_ref4) {
    var baseAnnotated = _ref4[0];
    return baseAnnotated;
  }).map(annotatedMemberToRemaining);

  if (maybePrefix) {
    var remaining = fromLeftBaseMembers.slice(1, 0);
    return [played, [maybePrefix].concat(remaining), [], metadata];
  }

  return [played, fromLeftBaseMembers, [], metadata];
}

function suffixFromPair(pair) {
  var baseMember = pair[0],
      againstMember = pair[1];

  if (!baseMember.isWindup) {
    return undefined;
  }

  if (!(againstMember != null && againstMember.isWindup)) {
    return undefined;
  }

  var metadata = getMetadata(baseMember.member);
  var baseMembers = allMembers(baseMember.member);
  var againstMembers = allMembers(againstMember.member);
  var zipped = zipMembersFromRight(baseMembers, againstMembers);

  var _matchesFromRight = matchesFromRight(zipped),
      matched = _matchesFromRight.matched,
      unmatchedFromRight = _matchesFromRight.unmatched;

  var lastUnmatchedFromRight = unmatchedFromRight[unmatchedFromRight.length - 1];
  var maybeSuffix = prefixFromPair(lastUnmatchedFromRight);
  var suffix = matched.map(annotatedMemberToPlayed);

  if (maybeSuffix) {
    var _remaining = baseMembers.slice(0, unmatchedFromRight.length - 1).map(annotatedMemberToRemaining);

    return [[], [].concat(_remaining, [maybeSuffix]), suffix, metadata];
  }

  var remaining = [].concat(unmatchedFromRight).map(function (_ref5) {
    var baseAnnotated = _ref5[0];
    return baseAnnotated;
  }).map(annotatedMemberToRemaining);
  return [[], remaining, suffix, metadata];
}

function annotatedMemberToRemaining(annotated) {
  if (annotated.isWindup && annotated.kind === MemberKind.Played) {
    // This shouldn't happen when this is used...
    return prev(playedWindupToSuffixed(annotated.member));
  }

  if (annotated.isWindup) {
    return windupToSuffixed(annotated.member);
  }

  return annotated.member;
}

function annotatedMemberToPlayed(annotated) {
  if (annotated.isWindup && annotated.kind === MemberKind.InProgress) {
    // This shouldn't happen when this is used...
    return fastForward(annotated.member);
  }

  return annotated.member;
}

function playedAndRemainingFromUnmatched(fromLeft, fromRight) {
  if (fromLeft.length === 0 || fromRight.length === 0) {
    return {
      played: [],
      remaining: []
    };
  }

  var _fromLeft$ = fromLeft[0],
      firstBaseFromLeft = _fromLeft$[0],
      firstAgainstFromLeft = _fromLeft$[1];
  var _fromRight = fromRight[fromRight.length - 1],
      firstBaseFromRight = _fromRight[0],
      firstAgainstFromRight = _fromRight[1];

  if (fromLeft.length === 1 && fromRight.length === 1 && firstBaseFromLeft.isWindup && firstAgainstFromLeft != null && firstAgainstFromLeft.isWindup) {
    var merged = mergeWindupsToSuffixedWindup(firstBaseFromLeft.member, firstAgainstFromLeft.member);

    if (isPlayedSuffixedWindup(merged)) {
      return {
        played: [merged],
        remaining: []
      };
    }

    return {
      played: [],
      remaining: [merged]
    };
  }

  var playedEls = fromLeft.filter(function (_ref6) {
    var baseAnnotated = _ref6[0];
    return baseAnnotated.kind === MemberKind.Played;
  });
  var played = playedEls.map(function (_ref7) {
    var baseAnnotated = _ref7[0];
    return baseAnnotated;
  }).map(function (baseAnnotated) {
    if (baseAnnotated.isWindup && baseAnnotated.kind === MemberKind.InProgress) {
      return fastForward$1(windupToSuffixed(baseAnnotated.member));
    }

    if (baseAnnotated.isWindup) {
      return playedWindupToSuffixed(baseAnnotated.member);
    }

    return baseAnnotated.member;
  });
  var shiftedLeft = fromLeft.slice(played.length);

  if (shiftedLeft.length === 0 && playedEls.length === 0) {
    return {
      played: played,
      remaining: []
    };
  }

  if (shiftedLeft.length === 0) {
    var shiftedFromRight = fromRight.slice(0, playedEls.length);
    var _shiftedFromRight = shiftedFromRight[shiftedFromRight.length - 1],
        baseFromRight = _shiftedFromRight[0],
        againstFromRight = _shiftedFromRight[1];

    var _maybeWindupSuffix = suffixFromPair([baseFromRight, againstFromRight]);

    if (_maybeWindupSuffix) {
      var playedWithoutSuffix = played.slice(0, -1);
      return {
        played: [].concat(playedWithoutSuffix, [fastForward$1(_maybeWindupSuffix)]),
        remaining: []
      };
    }

    return {
      played: played,
      remaining: []
    };
  }

  var _shiftedLeft$ = shiftedLeft[0],
      firstBaseFromShiftedLeft = _shiftedLeft$[0],
      firstAgainstFromShiftedLeft = _shiftedLeft$[1];
  var maybeWindupPrefix = prefixFromPair([firstBaseFromShiftedLeft, firstAgainstFromShiftedLeft]);
  var maybeWindupSuffix = suffixFromPair([firstBaseFromRight, firstAgainstFromRight]);
  var shiftedLeftBaseMembers = shiftedLeft.map(function (_ref8) {
    var baseAnnotated = _ref8[0];
    return baseAnnotated;
  }).map(annotatedMemberToRemaining);

  if (maybeWindupPrefix && maybeWindupSuffix) {
    var remaining = shiftedLeftBaseMembers.slice(1, -1);
    return {
      played: played,
      remaining: [maybeWindupPrefix].concat(remaining, [maybeWindupSuffix])
    };
  }

  if (!maybeWindupPrefix && maybeWindupSuffix) {
    var _remaining2 = shiftedLeftBaseMembers.slice(0, -1);

    return {
      played: played,
      remaining: [].concat(_remaining2, [maybeWindupSuffix])
    };
  }

  if (maybeWindupPrefix && !maybeWindupSuffix) {
    var _remaining3 = shiftedLeftBaseMembers.slice(1);

    return {
      played: played,
      remaining: [maybeWindupPrefix].concat(_remaining3)
    };
  }

  return {
    played: played,
    remaining: shiftedLeftBaseMembers
  };
}

function mergeWindupsToSuffixedWindup(baseWindup, againstWindup) {

  var metadata = getMetadata(baseWindup); // Get all the members and zip them up.

  var baseMembers = allMembers(baseWindup);
  var againstMembers = allMembers(againstWindup);
  var zipped = zipMembersFromLeft(baseMembers, againstMembers);

  var _matchesFromLeft2 = matchesFromLeft(zipped),
      matched = _matchesFromLeft2.matched,
      unmatchedFromLeft = _matchesFromLeft2.unmatched;

  var played = matched.map(function (member) {
    if (member.kind === MemberKind.InProgress && member.isWindup) {
      return playedWindupToSuffixed(fastForward(member.member));
    }

    if (member.kind === MemberKind.Played && member.isWindup) {
      return playedWindupToSuffixed(member.member);
    }

    return member.member;
  });
  var lengthToZipRight = baseMembers.length - played.length;
  var zippedRight = zipMembersFromRight(baseMembers, againstMembers).slice(0, lengthToZipRight);

  var _matchesFromRight2 = matchesFromRight(zippedRight),
      matchedFromRight = _matchesFromRight2.matched,
      unmatchedZippedFromRight = _matchesFromRight2.unmatched;

  var unmatchedZippedFromLeft = unmatchedFromLeft.slice(0, unmatchedZippedFromRight.length);

  var _playedAndRemainingFr = playedAndRemainingFromUnmatched(unmatchedZippedFromLeft, unmatchedZippedFromRight),
      playedFromUnmatched = _playedAndRemainingFr.played,
      remaining = _playedAndRemainingFr.remaining;

  var suffix = matchedFromRight.map(function (member) {
    if (member.kind === MemberKind.InProgress && member.isWindup) {
      return fastForward(member.member);
    }

    if (member.kind === MemberKind.Played && member.isWindup) {
      return member.member;
    }

    return member.member;
  });

  if (remaining.length === 0) {
    return [[].concat(played, playedFromUnmatched), suffix, metadata];
  }

  return [[].concat(played, playedFromUnmatched), remaining, suffix, metadata];
}
function getMetadata(windup) {
  return isPlayedWindup(windup) ? windup[1] : windup[2];
}
function lastPlayedSuffixedMember(windup) {
  if (isPlayedSuffixedWindup(windup)) {
    var _played2 = windup[0];
    var _last = _played2[_played2.length - 1];
    return _last;
  }

  var played = windup[0],
      remaining = windup[1];
  var playedFromRemaining = remaining.reduce(function (playedEl, member) {
    if (isSuffixedWindup(member)) {
      if (!isUnplayed$1(member)) {
        return lastPlayedSuffixedElement(member);
      }
    }

    return playedEl;
  }, undefined);

  if (playedFromRemaining) {
    return playedFromRemaining;
  }

  var last = played[played.length - 1];
  return last;
}
function lastPlayedSuffixedElement(windup) {
  var lastPlayed = lastPlayedSuffixedMember(windup);

  if (isPlayedSuffixedWindup(lastPlayed)) {
    return lastPlayedSuffixedElement(lastPlayed);
  }

  return lastPlayed;
}
function nextSuffixedElement(windup) {
  if (isPlayedSuffixedWindup(windup)) {
    return undefined;
  }

  var remaining = windup[1];
  var nextVal = remaining[0];

  if (isSuffixedWindup(nextVal)) {
    return nextSuffixedElement(nextVal);
  }

  return nextVal;
}
function unsuffixWindup(suffixed) {
  var played = suffixed[0],
      remaining = suffixed[1],
      suffix = suffixed[2],
      metadata = suffixed[3];
  var nextPlayed = played.map(function (member) {
    if (isPlayedSuffixedWindup(member)) {
      return unsuffixPlayedWindup(member);
    }

    return member;
  });
  var nextRemaining = remaining.map(function (member) {
    if (isSuffixedWindup(member)) {
      return unsuffixWindup(member);
    }

    return member;
  });
  var suffixAsMoreRemaining = suffix.map(function (member) {
    if (playedMemberIsWindup(member)) {
      return rewind(member);
    }

    return member;
  });
  return [nextPlayed, [].concat(nextRemaining, suffixAsMoreRemaining), metadata];
}
function unsuffixPlayedWindup(suffixed) {
  var played = suffixed[0],
      suffix = suffixed[1],
      metadata = suffixed[2];
  var nextPlayed = played.map(function (member) {
    if (isPlayedSuffixedWindup(member)) {
      return unsuffixPlayedWindup(member);
    }

    return member;
  });
  return [[].concat(nextPlayed, suffix), metadata];
}

function defaultGetPace(lastChar, nextChar) {
  switch (lastChar) {
    case "—":
    case "…":
      return 200;

    case ".":
    case ",":
      return 150;

    case "?":
    case "!":
      if (nextChar !== "!" && nextChar !== "?") {
        return 150;
      }

      return 20;

    case "-":
    case " ":
    case "\n":
      return 0;

    default:
      return 20;
  }
}

var Pace = function Pace(_ref) {
  var children = _ref.children;
  return createElement(Fragment, null, children);
};

function isPaceElement(element) {
  return element.type === Pace;
}
function isMsProp(props) {
  if ("ms" in props) {
    return true;
  }

  return false;
}
function isReverseMsProp(props) {
  if ("reverseMs" in props) {
    return true;
  }

  return false;
}
function isReverseGetPaceProp(props) {
  if ("reverseGetPace" in props) {
    return true;
  }

  return false;
}
function paceMetadataFromProps(props) {
  if (isMsProp(props)) {
    var ms = props.ms,
        reverseMs = props.reverseMs;
    return _extends({
      pace: function pace() {
        return ms;
      }
    }, reverseMs ? {
      reversePace: function reversePace() {
        return reverseMs;
      }
    } : {});
  }

  if (isReverseMsProp(props)) {
    var _ms = props.ms,
        _reverseMs = props.reverseMs;
    return _extends({
      reversePace: function reversePace() {
        return _reverseMs;
      }
    }, _ms ? {
      pace: function pace() {
        return _ms;
      }
    } : {});
  }

  if (isReverseGetPaceProp(props)) {
    return _extends({
      reversePace: props.getReversePace
    }, props.getPace ? {
      pace: props.getPace
    } : {});
  }

  return _extends({
    pace: props.getPace
  }, props.getReversePace ? {
    reversePace: props.getReversePace
  } : {});
}
function paceFromWindup(windup, parentPace) {
  if (isPlayedWindup(windup)) {
    return undefined;
  }

  var remaining = windup[1],
      metadata = windup[2];
  var firstRemaining = remaining[0];

  if (firstRemaining && memberIsWindup(firstRemaining)) {
    return paceFromWindup(firstRemaining, metadata.pace || parentPace);
  }

  return metadata.pace || parentPace;
}
function paceFromSuffixedWindup(windup, rewinding, parentPace) {
  if (isPlayedSuffixedWindup(windup)) {
    return undefined;
  }

  var remaining = windup[1],
      metadata = windup[3];
  var firstRemaining = remaining[0];
  var paceToUse = rewinding ? metadata.reversePace : metadata.pace;

  if (firstRemaining && isSuffixedWindup(firstRemaining)) {
    return paceFromSuffixedWindup(firstRemaining, rewinding, paceToUse || parentPace);
  }

  return paceToUse || parentPace;
}

var OnChar = function OnChar(_ref) {
  var children = _ref.children;
  return createElement(Fragment, null, children);
};

function isOnCharElement(element) {
  return element.type === OnChar;
} // Should return onChars from
// every just played member that is a windup
// any first remaining that is played

function onCharsFromWindup(windup) {
  if (isPlayedWindup(windup)) {
    return [];
  }

  var remaining = windup[1],
      metadata = windup[2];
  var lastPlayed = lastPlayedMember(windup);
  var firstRemaining = remaining[0];
  var onChars = [];

  if (lastPlayed && playedMemberIsWindup(lastPlayed)) {
    onChars.push.apply(onChars, onCharsFromWindup(lastPlayed));
  }

  if (firstRemaining && memberIsWindup(firstRemaining) && !isUnplayed(firstRemaining)) {
    onChars.push.apply(onChars, onCharsFromWindup(firstRemaining));
  }

  if (metadata.onChar) {
    onChars.push(metadata.onChar);
  }

  return onChars;
}
function onCharsFromSuffixedWindup(windup) {
  if (isPlayedSuffixedWindup(windup)) {
    return [];
  }

  var remaining = windup[1],
      metadata = windup[3];
  var lastPlayed = lastPlayedSuffixedMember(windup);
  var firstRemaining = remaining[0];
  var onChars = [];

  if (lastPlayed && isPlayedSuffixedWindup(lastPlayed)) {
    onChars.push.apply(onChars, onCharsFromSuffixedWindup(lastPlayed));
  }

  if (firstRemaining && isSuffixedWindup(firstRemaining) && !isUnplayed$1(firstRemaining)) {
    onChars.push.apply(onChars, onCharsFromSuffixedWindup(firstRemaining));
  }

  if (metadata.onChar) {
    onChars.push(metadata.onChar);
  }

  return onChars;
}

function initWindupState(windup) {
  return {
    windup: windup,
    didFinishOnce: false
  };
}

function windupReducer(state, action) {
  switch (action.type) {
    case "replace":
      return initWindupState(action.windup);

    case "next":
      if (isPlayedWindup(state.windup)) {
        return state;
      }

      return _extends({}, state, {
        windup: next(state.windup)
      });

    case "rewind":
      return {
        windup: rewind(state.windup),
        didFinishOnce: false
      };

    case "fast-forward":
      if (isPlayedWindup(state.windup)) {
        return state;
      }

      return _extends({}, state, {
        windup: fastForward(state.windup)
      });

    case "finish":
      return _extends({}, state, {
        didFinishOnce: true
      });

    default:
      return state;
  }
}

function useWindup(windupInit, options) {
  var _React$useReducer = useReducer(windupReducer, windupInit, initWindupState),
      _React$useReducer$ = _React$useReducer[0],
      windup = _React$useReducer$.windup,
      didFinishOnce = _React$useReducer$.didFinishOnce,
      dispatch = _React$useReducer[1];

  var timeoutRef = useRef(null);
  var windupIsFinished = isFinished(windup);
  var skip = useCallback(function () {
    if (!windupIsFinished) {
      if (timeoutRef.current) {
        clearTimeout(timeoutRef.current);
      }

      dispatch({
        type: "fast-forward"
      });
    }
  }, [windupIsFinished]);
  var rewind = useCallback(function () {
    if (timeoutRef.current) {
      clearTimeout(timeoutRef.current);
    }

    dispatch({
      type: "rewind"
    });
  }, []); // If windup arg changes, we should reset

  useEffect(function () {
    dispatch({
      type: "replace",
      windup: windupInit
    });
  }, [windupInit]); // When the windup changes, onChar should fire

  useEffect(function () {
    var onChars = onCharsFromWindup(windup);
    var lastEl = lastPlayedElement(windup);

    if (onChars.length > 0 && lastEl) {
      onChars.forEach(function (onChar) {
        onChar(lastEl, false);
      });
    }
  }, [windup]); // If windup finishes, the onFinished should fire

  useEffect(function () {
    // Put this in a new context so that the windup finishes visually before firing this
    if (didFinishOnce === false && windupIsFinished) {
      var timeout = setTimeout(function () {
        if (options.onFinished) {
          options.onFinished();
        }

        dispatch({
          type: "finish"
        });
      }, 0);
      return function () {
        clearTimeout(timeout);
      };
    }
  }, [didFinishOnce, windupIsFinished, options]); // the windup effect itself

  useEffect(function () {
    if (!windupIsFinished) {
      var getPace = paceFromWindup(windup) || defaultGetPace;
      var lastEl = lastPlayedElement(windup);
      var nextEl = nextElement(windup);
      timeoutRef.current = setTimeout(function () {
        dispatch({
          type: "next"
        });
      }, lastEl ? getPace(lastEl, nextEl) : 0);
      return function () {
        if (timeoutRef.current) {
          clearTimeout(timeoutRef.current);
        }
      };
    }
  }, [windup, windupIsFinished]);
  return {
    windup: windup,
    skip: skip,
    rewind: rewind,
    isFinished: windupIsFinished
  };
}

function renderStringWindup(windup) {
  var played = playedElements(windup);
  var inner = played.reduce(function (acc, playedEl) {
    if (typeof playedEl === "string") {
      return acc + playedEl;
    }

    return acc + renderStringWindup(playedEl);
  }, "");
  return inner;
}
var VOID_TAGS = ["area", "base", "br", "col", "command", "embed", "hr", "img", "input", "keygen", "link", "meta", "param", "source", "track", "wbr"];
function renderChildrenWindup(windup) {
  var metadata = isPlayedWindup(windup) ? windup[1] : windup[2];
  var played = playedElements(windup);
  var Outer = metadata.element || Fragment;

  if (metadata.props && Object.keys(metadata.props).includes("children")) {
    return createElement(Outer, Object.assign({
      key: metadata.key
    }, metadata.props));
  }

  var inner = played.reduce(function (acc, playedEl) {
    if (typeof playedEl === "string") {
      var accButLast = acc.slice(0, acc.length - 1);
      var last = acc[acc.length - 1];
      return last && typeof last === "string" ? [].concat(accButLast, [last + playedEl]) : [].concat(acc, [playedEl]);
    }

    return [].concat(acc, [renderChildrenWindup(playedEl)]);
  }, []);

  if (typeof metadata.element === "string" && VOID_TAGS.includes(metadata.element)) {
    return createElement(Outer, Object.assign({
      key: metadata.key
    }, metadata.props));
  }

  if (inner.every(function (node) {
    return typeof node === "string";
  })) {
    return createElement(Outer, Object.assign({
      key: metadata.key
    }, metadata.props), inner.join(""));
  }

  return createElement(Outer, Object.assign({
    key: metadata.key
  }, metadata.props), inner);
}

function useWindupString(text, options) {
  if (options === void 0) {
    options = {};
  }

  var windupInit = useMemo(function () {
    return windupFromString(text, options); // We can omit options as this is used as an initialisation value and options will not change it
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [text]);
  var maybeSkippedWindupInit = useMemo(function () {
    if (options.skipped) {
      return fastForward(windupInit);
    }

    return windupInit;
  }, [windupInit, options.skipped]);

  var _useWindup = useWindup(maybeSkippedWindupInit, options),
      windup = _useWindup.windup,
      skip = _useWindup.skip,
      rewind = _useWindup.rewind,
      isFinished = _useWindup.isFinished;

  var rendered = renderStringWindup(windup);
  useDebugValue(rendered);
  return [rendered, {
    skip: skip,
    rewind: rewind,
    isFinished: isFinished
  }];
}

var Pause = function Pause() {
  return null;
};

function isPauseElement(element) {
  return element.type === Pause;
}

var WindupContext = /*#__PURE__*/createContext({
  skip: function skip() {
    console.warn("Tried to use the useSkip hook outside of a WindupChildren component!!");
  },
  rewind: function rewind() {
    console.warn("Tried to use the useRewind hook outside of a WindupChildren component!");
  },
  isFinished: false
});
function useSkip() {
  var _React$useContext = useContext(WindupContext),
      skip = _React$useContext.skip;

  return skip;
}
function useRewind() {
  var _React$useContext2 = useContext(WindupContext),
      rewind = _React$useContext2.rewind;

  return rewind;
}
function useIsFinished() {
  var _React$useContext3 = useContext(WindupContext),
      isFinished = _React$useContext3.isFinished;

  return isFinished;
}
function reduceWindupArgs(prevArgs, children) {
  if (typeof children === "string") {
    return [].concat(prevArgs, children.split(""));
  }

  if (typeof children === "number") {
    return [].concat(prevArgs, children.toString().split(""));
  }

  if (!isValidElement(children)) {
    return prevArgs;
  }

  var _children$props = children.props,
      childrenChildren = _children$props.children,
      restProps = _objectWithoutPropertiesLoose(_children$props, ["children"]);

  var paceMetaData = isPaceElement(children) ? paceMetadataFromProps(children.props) : {};
  var onCharMetaData = isOnCharElement(children) ? {
    onChar: children.props.fn
  } : {};
  var keyProp = children.key ? {
    key: children.key
  } : {};

  if (isPauseElement(children)) {
    return [].concat(prevArgs, [windupFromString(" ", _extends({
      element: Fragment
    }, keyProp, {
      props: {
        children: undefined
      },
      pace: function pace() {
        return children.props.ms;
      }
    }))]);
  }

  if (childrenChildren === undefined) {
    return [].concat(prevArgs, [windupFromString(" ", _extends({
      element: children.type,
      props: _extends({}, restProps, {
        children: undefined
      })
    }, keyProp, paceMetaData, onCharMetaData))]);
  }

  if (typeof childrenChildren === "string") {
    return [].concat(prevArgs, [windupFromString(childrenChildren, _extends({
      element: children.type,
      props: restProps
    }, keyProp, paceMetaData, onCharMetaData))]);
  }

  if (childrenChildren instanceof Function) {
    return [].concat(prevArgs, [windupFromString(" ", _extends({
      element: children.type,
      props: _extends({
        children: childrenChildren
      }, restProps)
    }, keyProp, paceMetaData, onCharMetaData))]);
  }

  var newArgs = Children.toArray(childrenChildren).reduce(reduceWindupArgs, []);
  var argsWithMetadata = newArgs.map(function (member) {
    if (memberIsWindup(member)) {
      var played = member[0],
          remaining = member[1],
          metadata = member[2];
      return [played, remaining, _extends({}, paceMetaData, onCharMetaData, metadata)];
    }

    return member;
  });
  return [].concat(prevArgs, [newWindup(argsWithMetadata, _extends({
    element: children.type,
    props: restProps
  }, keyProp, paceMetaData, onCharMetaData))]);
}

function buildKeyString(children) {
  if (children === null || children === undefined) {
    return "";
  }

  var keys = Children.map(children, function (child) {
    if (typeof child === "string") {
      return child;
    }

    if (typeof child === "number") {
      return child.toString();
    }

    if (isValidElement(child)) {
      return "#" + (child.key || "") + "<" + buildKeyString(child.props.children) + ">";
    }

    return "";
  });

  if (!keys) {
    return "";
  }

  return keys.join(",");
}

function useChildrenMemo(factory, children) {
  // Omitting children in favour of using a key instead
  // eslint-disable-next-line react-hooks/exhaustive-deps
  var memoChildren = useMemo(factory, [buildKeyString(children)]);
  return memoChildren;
}

var WindupChildren = function WindupChildren(_ref) {
  var children = _ref.children,
      onFinished = _ref.onFinished,
      skipped = _ref.skipped;
  var windupInit = useChildrenMemo(function () {
    return newWindup(Children.toArray(children).reduce(reduceWindupArgs, []), {
      element: undefined
    });
  }, children);
  var maybeSkippedWindupInit = useMemo(function () {
    if (skipped) {
      return fastForward(windupInit);
    }

    return windupInit;
  }, [windupInit, skipped]);

  var _useWindup = useWindup(maybeSkippedWindupInit, {
    onFinished: onFinished
  }),
      windup = _useWindup.windup,
      skip = _useWindup.skip,
      rewind = _useWindup.rewind,
      isFinished = _useWindup.isFinished;

  return createElement(WindupContext.Provider, {
    value: {
      skip: skip,
      rewind: rewind,
      isFinished: isFinished
    }
  }, renderChildrenWindup(windup));
};

function wrapChildren(children, Wrapper) {
  if (typeof children === "string") {
    return children.split("").map(function (_char, i) {
      return _char === "\n" ? _char : createElement(Wrapper, {
        key: _char + "-" + i
      }, _char);
    });
  }

  if (typeof children === "number") {
    return children.toString().split("").map(function (_char2) {
      return createElement(Wrapper, null, _char2);
    });
  }

  if (!isValidElement(children)) {
    return createElement(Fragment, null);
  }

  if (children.props.children) {
    var Outer = children.type;

    if (children.props.children instanceof Function) {
      return createElement(Outer, Object.assign({
        key: children.key
      }, children.props));
    }

    return createElement(Outer, Object.assign({
      key: children.key
    }, children.props), Children.map(children.props.children, function (ch) {
      return wrapChildren(ch, Wrapper);
    }));
  }

  return children;
}

var CharWrapper = function CharWrapper(_ref) {
  var children = _ref.children,
      element = _ref.element;
  return createElement(Fragment, null, Children.map(children, function (ch) {
    return wrapChildren(ch, element);
  }));
};

function makeGetDescriptorsOfChildren(defaultFontStyle) {
  return function (descriptors, children) {
    if (typeof children === "string") {
      return [].concat(descriptors, [{
        text: children,
        font: defaultFontStyle
      }]);
    }

    if (typeof children === "number") {
      return [].concat(descriptors, [{
        text: children.toString(),
        font: defaultFontStyle
      }]);
    }

    if (!isValidElement(children)) {
      return descriptors;
    }

    if (isStyledTextElement(children) && typeof children.props.children === "string") {
      return [].concat(descriptors, [{
        text: children.props.children,
        font: children.props.fontStyle
      }]);
    }

    if (isStyledTextElement(children)) {
      return [].concat(descriptors, Children.toArray(children.props.children).reduce(makeGetDescriptorsOfChildren(children.props.fontStyle), []));
    }

    return [].concat(descriptors, Children.toArray(children.props.children).reduce(makeGetDescriptorsOfChildren(defaultFontStyle), []));
  };
}

function reinsertStringsIntoChildren(_ref, children) {
  var accChildren = _ref[0],
      accStrings = _ref[1];

  if (typeof children === "string" || typeof children === "number") {
    var firstString = accStrings[0],
        restStrings = accStrings.slice(1);
    return [[].concat(accChildren, [firstString]), restStrings];
  }

  if (!isValidElement(children)) {
    return [accChildren, accStrings];
  }

  var _React$Children$toArr = Children.toArray(children.props.children).reduce(reinsertStringsIntoChildren, [[], accStrings]),
      subChildrenAcc = _React$Children$toArr[0],
      subStringsAcc = _React$Children$toArr[1];

  return [[].concat(accChildren, [cloneElement(children, {
    children: subChildrenAcc
  })]), subStringsAcc];
}

function isStyledTextElement(element) {
  return element.type === StyledText;
}

function StyledText(_ref2) {
  var children = _ref2.children;
  return createElement(Fragment, null, children);
}

var Linebreaker = function Linebreaker(_ref3) {
  var children = _ref3.children,
      fontStyle = _ref3.fontStyle,
      width = _ref3.width;
  // CAVEATS:
  // fontStyle must match the font style of the characters inside
  // non-character elements must not add width to the line.
  // must be used OUTSIDE of WindupChildren component
  var childrenArray = Children.toArray(children);
  var descriptors = childrenArray.reduce(makeGetDescriptorsOfChildren(fontStyle), []);
  var transformedStrings = breakLines(descriptors, width, fontStyle);

  var _childrenArray$reduce = childrenArray.reduce(reinsertStringsIntoChildren, [[], transformedStrings]),
      transformedChildren = _childrenArray$reduce[0];

  return createElement("div", {
    style: {
      whiteSpace: "pre"
    }
  }, transformedChildren);
};

var Effect = function Effect(_ref) {
  var fn = _ref.fn;
  useEffect(function () {
    fn(); // We can safely omit fn from dependencies as if its value changes
    // the whole windup will be re-rendered anyway
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, []);
  return null;
};

function textFromChildren(children) {
  var text = Children.map(children, function (ch) {
    if (typeof ch === "string") {
      return ch;
    }

    if (typeof ch === "number") {
      return ch.toString();
    }

    if (!isValidElement(ch)) {
      return "";
    }

    if (ch.props.children) {
      return textFromChildren(ch.props.children);
    }

    return "";
  });

  if (!text) {
    return "";
  }

  return text.join("");
}

function renderSuffixedStringWindup(windup) {
  var played = playedElementsWithSuffix(windup);
  var inner = played.reduce(function (acc, playedEl) {
    if (typeof playedEl === "string") {
      return acc + playedEl;
    }

    return acc + renderSuffixedStringWindup(playedEl);
  }, "");
  return inner;
}
function renderSuffixedChildrenWindup(windup) {
  var metadata = isPlayedSuffixedWindup(windup) ? windup[2] : windup[3];
  var played = playedElementsWithSuffix(windup);
  var Outer = metadata.element || Fragment;

  if (metadata.props && Object.keys(metadata.props).includes("children")) {
    return createElement(Outer, Object.assign({
      key: metadata.key
    }, metadata.props));
  }

  var inner = played.reduce(function (acc, playedEl) {
    if (typeof playedEl === "string") {
      return [].concat(acc, [playedEl]);
    }

    return [].concat(acc, [renderSuffixedChildrenWindup(playedEl)]);
  }, []);

  if (typeof metadata.element === "string" && VOID_TAGS.includes(metadata.element)) {
    return createElement(Outer, Object.assign({
      key: metadata.key
    }, metadata.props));
  }

  if (inner.every(function (node) {
    return typeof node === "string";
  })) {
    return createElement(Outer, Object.assign({
      key: metadata.key
    }, metadata.props), inner.join(""));
  }

  return createElement(Outer, Object.assign({
    key: metadata.key
  }, metadata.props), inner);
}

function usePrevious(value) {
  var ref = useRef();
  useEffect(function () {
    ref.current = value;
  }, [value]);
  return ref.current;
}

function isRewindupState(state) {
  if ("toRewind" in state) {
    return true;
  }

  return false;
}

function isRewinding(state) {
  return isRewindupState(state) ? playedSize(state.toRewind) > playedSize(state.toResume) : false;
}

function getActiveWindup(state) {
  return isRewindupState(state) ? isRewinding(state) ? state.toRewind : state.toResume : state.toPlay;
}

function getNextBaseWindup(state) {
  if (!isRewindupState(state)) {
    return isPlayedSuffixedWindup(state.toPlay) ? unsuffixPlayedWindup(state.toPlay) : unsuffixWindup(state.toPlay);
  }

  if (isRewinding(state)) {
    return isPlayedSuffixedWindup(state.toRewind) ? unsuffixPlayedWindup(state.toRewind) : unsuffixWindup(state.toRewind);
  }

  return isPlayedSuffixedWindup(state.toResume) ? unsuffixPlayedWindup(state.toResume) : unsuffixWindup(state.toResume);
}

function rewindupReducer(state, action) {
  switch (action.type) {
    case "start-rewindup":
      {
        var toRewind = mergeWindupsToSuffixedWindup(action.baseWindup, action.againstWindup);
        var toResume = mergeWindupsToSuffixedWindup(action.againstWindup, action.baseWindup);
        return {
          toRewind: toRewind,
          toResume: toResume
        };
      }

    case "proceed":
      {
        if (!isRewindupState(state)) {
          if (isPlayedSuffixedWindup(state.toPlay)) {
            return state;
          }

          return {
            toPlay: next$1(state.toPlay)
          };
        }

        if (isRewinding(state)) {
          var nextToRewind = prev(state.toRewind);
          return _extends({}, state, {
            toRewind: nextToRewind
          });
        }

        var nextToResume = next$1(state.toResume);
        return _extends({}, state, {
          toResume: nextToResume
        });
      }

    case "fast-forward":
      {
        return {
          toPlay: fastForward$1(getActiveWindup(state))
        };
      }

    default:
      return state;
  }
}

function useRewindup(windupInit, options) {
  var _React$useReducer = useReducer(rewindupReducer, {
    toPlay: isPlayedWindup(windupInit) ? playedWindupToSuffixed(windupInit) : windupToSuffixed(windupInit)
  }),
      state = _React$useReducer[0],
      dispatch = _React$useReducer[1]; // When the windup changes, onChar should fire


  useEffect(function () {
    if (!state) {
      return;
    }

    var windupToGetOnCharFrom = getActiveWindup(state);
    var onChars = onCharsFromSuffixedWindup(windupToGetOnCharFrom);
    var lastEl = lastPlayedSuffixedElement(windupToGetOnCharFrom);

    if (onChars.length > 0 && lastEl) {
      onChars.forEach(function (onChar) {
        onChar(lastEl, isRewindupState(state));
      });
    }
  }, [state]);
  var timeoutRef = useRef(null); // If windup finishes, the onFinished should fire

  var windupIsFinished = isPlayedSuffixedWindup(getActiveWindup(state));
  var didFinishOnce = useRef(false);
  useEffect(function () {
    // Put this in a new context so that the windup finishes visually before firing this
    if (windupIsFinished && !didFinishOnce.current) {
      var timeout = setTimeout(function () {
        if (options.onFinished) {
          options.onFinished();
        }

        didFinishOnce.current = true;
      }, 0);
      return function () {
        clearTimeout(timeout);
      };
    }
  }, [windupIsFinished, options.onFinished]); // Proceed the windup

  useEffect(function () {
    var isPlayed = isPlayedSuffixedWindup(isRewindupState(state) ? state.toResume : state.toPlay);
    var isRewindDone = isRewindupState(state) ? playedSize(state.toResume) >= playedSize(state.toRewind) : true;

    if (!state || isPlayed && isRewindDone) {
      return;
    }

    var windupToGetPaceFrom = getActiveWindup(state);
    var getPace = paceFromSuffixedWindup(windupToGetPaceFrom, isRewinding(state)) || defaultGetPace;
    var lastEl = lastPlayedSuffixedElement(windupToGetPaceFrom);
    var nextEl = nextSuffixedElement(windupToGetPaceFrom);
    var timeout = lastEl ? getPace(lastEl, nextEl) : 0;
    timeoutRef.current = setTimeout(function () {
      dispatch({
        type: "proceed"
      });
    }, timeout);
    return function () {
      if (timeoutRef.current) {
        clearTimeout(timeoutRef.current);
      }
    };
  }, [state]);
  var prevWindupInit = usePrevious(windupInit);
  useEffect(function () {
    didFinishOnce.current = false;

    if (prevWindupInit === undefined) {
      return;
    }

    if (prevWindupInit === windupInit) {
      return;
    }

    if (options.onRewindStart) {
      options.onRewindStart();
    }

    dispatch({
      type: "start-rewindup",
      baseWindup: getNextBaseWindup(state),
      againstWindup: windupInit
    });
  }, [windupInit, prevWindupInit]);
  var skip = useCallback(function () {
    dispatch({
      type: "fast-forward"
    });
  }, []);
  var rewind = useCallback(function () {
    dispatch({
      type: "start-rewindup",
      baseWindup: newWindup([], {}),
      againstWindup: windupInit
    });
    didFinishOnce.current = false;
  }, []);
  return {
    windup: getActiveWindup(state),
    skip: skip,
    rewind: rewind,
    isFinished: windupIsFinished
  };
}

function useRewindupString(text, options) {
  if (options === void 0) {
    options = {};
  }

  var windupInit = useMemo(function () {
    return windupFromString(text, options); // We can omit options as this is used as an initialisation value and options will not change it
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [text]);
  var maybeSkippedWindupInit = useMemo(function () {
    if (options.skipped) {
      return fastForward(windupInit);
    }

    return windupInit;
  }, [windupInit, options.skipped]);

  var _useRewindup = useRewindup(maybeSkippedWindupInit, options),
      windup = _useRewindup.windup,
      skip = _useRewindup.skip,
      rewind = _useRewindup.rewind,
      isFinished = _useRewindup.isFinished;

  var rendered = renderSuffixedStringWindup(windup);
  useDebugValue(rendered);
  return [rendered, {
    skip: skip,
    rewind: rewind,
    isFinished: isFinished
  }];
}

function RewindupChildren(_ref) {
  var children = _ref.children,
      onFinished = _ref.onFinished,
      onRewindStart = _ref.onRewindStart,
      skipped = _ref.skipped;
  var windupInit = useChildrenMemo(function () {
    if (children === undefined) {
      return newWindup([], {
        element: undefined
      });
    }

    return newWindup(Children.toArray(children).reduce(reduceWindupArgs, []), {
      element: undefined
    });
  }, children);
  var maybeSkippedWindupInit = useMemo(function () {
    if (skipped) {
      return fastForward(windupInit);
    }

    return windupInit;
  }, [windupInit, skipped]);

  var _useRewindup = useRewindup(maybeSkippedWindupInit, {
    onFinished: onFinished,
    onRewindStart: onRewindStart
  }),
      windup = _useRewindup.windup,
      skip = _useRewindup.skip,
      rewind = _useRewindup.rewind,
      isFinished = _useRewindup.isFinished;

  return createElement(WindupContext.Provider, {
    value: {
      skip: skip,
      rewind: rewind,
      isFinished: isFinished
    }
  }, renderSuffixedChildrenWindup(windup));
}

export { CharWrapper, Effect, Linebreaker, OnChar, Pace, Pause, RewindupChildren, StyledText, WindupChildren, defaultGetPace, textFromChildren, useIsFinished, useRewind, useRewindupString, useSkip, useWindupString };
//# sourceMappingURL=windups.esm.js.map
