import * as React from "react";
import {
  isPlayedSuffixedWindup,
  isSuffixedWindup,
  lastPlayedSuffixedMember,
  PlayedSuffixedWindup,
  SuffixedWindup,
  isUnplayed as isUnplayedSuffixed,
} from "../SuffixedWindup";
import {
  memberIsWindup,
  Windup,
  lastPlayedMember,
  isUnplayed,
  PlayedWindup,
  isPlayedWindup,
  playedMemberIsWindup,
} from "../Windup";
import { SuffixedHookMetadata } from "./useRewindup";
import { HookMetadata } from "./useWindup";

type OnCharProps = {
  fn: (char: string, reversing?: boolean) => void;
};

const OnChar: React.FC<OnCharProps> = ({ children }) => {
  return <>{children}</>;
};

export function isOnCharElement(
  element: React.ReactElement
): element is React.ReactElement<OnCharProps> {
  return element.type === OnChar;
}

// Should return onChars from
// every just played member that is a windup
// any first remaining that is played
export function onCharsFromWindup<MetadataType extends HookMetadata>(
  windup: Windup<string, MetadataType> | PlayedWindup<string, MetadataType>
): ((char: string, reversing?: boolean) => void)[] {
  if (isPlayedWindup(windup)) {
    return [];
  }

  const [, remaining, metadata] = windup;
  const lastPlayed = lastPlayedMember(windup);
  const [firstRemaining] = remaining;

  let onChars = [];

  if (lastPlayed && playedMemberIsWindup(lastPlayed)) {
    onChars.push(...onCharsFromWindup(lastPlayed));
  }

  if (
    firstRemaining &&
    memberIsWindup(firstRemaining) &&
    !isUnplayed(firstRemaining)
  ) {
    onChars.push(...onCharsFromWindup(firstRemaining));
  }

  if (metadata.onChar) {
    onChars.push(metadata.onChar);
  }

  return onChars;
}

export function onCharsFromSuffixedWindup<
  MetadataType extends SuffixedHookMetadata
>(
  windup:
    | SuffixedWindup<string, MetadataType>
    | PlayedSuffixedWindup<string, MetadataType>
): ((char: string, reversing?: boolean) => void)[] {
  if (isPlayedSuffixedWindup(windup)) {
    return [];
  }

  const [, remaining, , metadata] = windup;
  const lastPlayed = lastPlayedSuffixedMember(windup);
  const [firstRemaining] = remaining;

  let onChars = [];

  if (lastPlayed && isPlayedSuffixedWindup(lastPlayed)) {
    onChars.push(...onCharsFromSuffixedWindup(lastPlayed));
  }

  if (
    firstRemaining &&
    isSuffixedWindup(firstRemaining) &&
    !isUnplayedSuffixed(firstRemaining)
  ) {
    onChars.push(...onCharsFromSuffixedWindup(firstRemaining));
  }

  if (metadata.onChar) {
    onChars.push(metadata.onChar);
  }

  return onChars;
}

export default OnChar;
