import * as React from "react";
import { fastForward, newWindup } from "../Windup";
import { renderSuffixedChildrenWindup } from "./renderSuffixedWindup";
import useRewindup from "./useRewindup";
import {
  ChildrenMetadata,
  reduceWindupArgs,
  useChildrenMemo,
  WindupContext,
} from "./WindupChildren";

type RewindupChildrenProps = {
  children: React.ReactNode;
  onRewindStart?: () => void;
  onFinished?: () => void;
  skipped?: boolean;
};

export default function RewindupChildren({
  children,
  onFinished,
  onRewindStart,
  skipped,
}: RewindupChildrenProps) {
  const windupInit = useChildrenMemo(() => {
    if (children === undefined) {
      return newWindup<string, ChildrenMetadata>([], { element: undefined });
    }

    return newWindup<string, ChildrenMetadata>(
      React.Children.toArray(children).reduce(reduceWindupArgs, []),
      { element: undefined }
    );
  }, children);

  const maybeSkippedWindupInit = React.useMemo(() => {
    if (skipped) {
      return fastForward(windupInit);
    }

    return windupInit;
  }, [windupInit, skipped]);

  const { windup, skip, rewind, isFinished } = useRewindup(
    maybeSkippedWindupInit,
    {
      onFinished,
      onRewindStart,
    }
  );

  return (
    <WindupContext.Provider
      value={{
        skip,
        rewind,
        isFinished,
      }}
    >
      {renderSuffixedChildrenWindup(windup)}
    </WindupContext.Provider>
  );
}
