import * as React from "react";
import {
  isPlayedSuffixedWindup,
  playedElementsWithSuffix,
  PlayedSuffixedWindup,
  SuffixedWindup,
} from "../SuffixedWindup";
import { VOID_TAGS } from "./renderWindup";
import { StringMetadata } from "./useWindupString";
import { ChildrenMetadata } from "./WindupChildren";

type SuffixedStringWindup = SuffixedWindup<string, StringMetadata>;
type SuffixedPlayedStringWindup = PlayedSuffixedWindup<string, StringMetadata>;

export function renderSuffixedStringWindup(
  windup: SuffixedStringWindup | SuffixedPlayedStringWindup
): string {
  const played = playedElementsWithSuffix(windup);

  const inner = played.reduce((acc: string, playedEl) => {
    if (typeof playedEl === "string") {
      return acc + playedEl;
    }

    return acc + renderSuffixedStringWindup(playedEl);
  }, "");

  return inner;
}

export type SuffixedChildrenWindup = SuffixedWindup<string, ChildrenMetadata>;
export type SuffixedPlayedChildrenWindup = PlayedSuffixedWindup<
  string,
  ChildrenMetadata
>;

export function renderSuffixedChildrenWindup(
  windup: SuffixedChildrenWindup | SuffixedPlayedChildrenWindup
): React.ReactNode {
  const metadata = isPlayedSuffixedWindup(windup) ? windup[2] : windup[3];
  const played = playedElementsWithSuffix(windup);

  const Outer = metadata.element || React.Fragment;

  if (metadata.props && Object.keys(metadata.props).includes("children")) {
    return <Outer key={metadata.key} {...metadata.props} />;
  }

  const inner = played.reduce((acc: React.ReactNode[], playedEl) => {
    if (typeof playedEl === "string") {
      return [...acc, playedEl];
    }

    return [...acc, renderSuffixedChildrenWindup(playedEl)];
  }, []);

  if (
    typeof metadata.element === "string" &&
    VOID_TAGS.includes(metadata.element)
  ) {
    return <Outer key={metadata.key} {...metadata.props} />;
  }

  if (inner.every((node) => typeof node === "string")) {
    return (
      <Outer key={metadata.key} {...metadata.props}>
        {inner.join("")}
      </Outer>
    );
  }

  return (
    <Outer key={metadata.key} {...metadata.props}>
      {inner}
    </Outer>
  );
}
