import * as React from "react";
import { windupFromString, fastForward } from "../Windup";
import { renderSuffixedStringWindup } from "./renderSuffixedWindup";
import useRewindup, { SuffixedHookMetadata } from "./useRewindup";
import { WindupOptions } from "./useWindupString";

export type RewindupOptions = {
  reversePace?: (char: string) => number;
} & WindupOptions;

export default function useRewindupString(
  text: string,
  options: WindupOptions = {}
): [
  string,
  {
    skip: () => void;
    rewind: () => void;
    isFinished: boolean;
  }
] {
  const windupInit = React.useMemo(() => {
    return windupFromString<SuffixedHookMetadata>(text, options);
    // We can omit options as this is used as an initialisation value and options will not change it
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [text]);
  
  const maybeSkippedWindupInit = React.useMemo(() => {
    if (options.skipped) {
      return fastForward(windupInit);
    }

    return windupInit;
  }, [windupInit, options.skipped]);

  const { windup, skip, rewind, isFinished } = useRewindup(maybeSkippedWindupInit, options);

  let rendered = renderSuffixedStringWindup(windup);

  React.useDebugValue(rendered);

  return [rendered, { skip, rewind, isFinished }];
}
