import * as React from "react";
import { windupFromString, Windup, fastForward } from "../Windup";
import useWindup, { HookMetadata } from "./useWindup";
import { renderStringWindup } from "./renderWindup";

export type WindupOptions = {
  onChar?: (char: string, reversing?: boolean) => void;
  onFinished?: () => void;
  pace?: (char: string) => number;
  skipped?: boolean;
};

export type StringMetadata = HookMetadata;

export type StringWindup = Windup<string, StringMetadata>;

export default function useWindupString(
  text: string,
  options: WindupOptions = {}
): [
  string,
  {
    skip: () => void;
    rewind: () => void;
    isFinished: boolean;
  }
] {
  const windupInit = React.useMemo(() => {
    return windupFromString<StringMetadata>(text, options);
    // We can omit options as this is used as an initialisation value and options will not change it
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [text]);
  
  const maybeSkippedWindupInit = React.useMemo(() => {
    if (options.skipped) {
      return fastForward(windupInit);
    }

    return windupInit;
  }, [windupInit, options.skipped]);

  const { windup, skip, rewind, isFinished } = useWindup(maybeSkippedWindupInit, options);

  const rendered = renderStringWindup(windup);
  React.useDebugValue(rendered);

  return [rendered, { skip, rewind, isFinished }];
}
