import {
  isPlayedSuffixedWindup,
  mergeWindupsToSuffixedWindup,
  PlayedSuffixedWindup,
  SuffixedWindup,
} from "../SuffixedWindup";
import { PlayedWindup, Windup, windupFromString, fastForward } from "../Windup";
import { deepPair } from "./util";

type PairElement<E, M> = Windup<E, M> | PlayedWindup<E, M>;
type Pair<E, M> = [PairElement<E, M>, PairElement<E, M>];
type Expected<E, M> = SuffixedWindup<E, M> | PlayedSuffixedWindup<E, M>;

type Vector<E, M> = [string, Pair<E, M>, Expected<E, M>];

const finishedSentence = fastForward(
  windupFromString("windups can't intelligently diff contents", {
    m: true,
  })
);
const unfinishedSentence = windupFromString(
  "windups can definitely diff their contents",
  {
    m: true,
  }
);

const vectors: Vector<string, { m: boolean }>[] = [
  [
    "Finished as base",
    [finishedSentence, unfinishedSentence],
    [
      "windups can't intelligently diff".split(""),
      " contents".split(""),
      { m: true },
    ],
  ],
  [
    "Unfinished as base",
    [unfinishedSentence, finishedSentence],
    [
      "windups can".split(""),
      " definitely diff their".split(""),
      " contents".split(""),
      { m: true },
    ],
  ],
  [
    "Deep",
    [deepPair.base, deepPair.against],
    [
      ["H", "i", "!", " "],
      [
        [
          ["W", "h", "a", "t", " "],
          [ "a", " ", "n", "i", "c", "e"],
          [" ", "r", "o", "o", "m", "y"],
          {
            m: false,
          },
        ],
      ],
      " place".split(""),
      { m: true },
    ],
  ],
  [
    "Deep played",
    [fastForward(deepPair.base), deepPair.against],
    [
      [
        "H",
        "i",
        "!",
        " ",
        [
          ["W", "h", "a", "t", " ", "a", " ", "n", "i", "c", "e"],
          [" ", "r", "o", "o", "m", "y"],
          {
            m: false,
          },
        ],
      ],
      " place".split(""),
      { m: true },
    ],
  ],
  [
    "Deep against played",
    [deepPair.against, fastForward(deepPair.base)],
    [
      ["H", "i", "!", " "],
      [
        [
          ["W", "h", "a", "t", " ",],
          ["i", "s", " ", "t", "h", "i", "s"],
          [" ", "r", "o", "o", "m", "y"],
          {
            m: false,
          },
        ],
      ],
      " place".split(""),
      { m: true },
    ],
  ],
  ["Partial played",
  [[['A'], [' ', 'c', 'a', 't'], { m: true}], [[], ['A', ' ', 'c', 'o', 'b'], { m: true}]],
  [
    ['A', ' ', 'c'], 
    ['a', 't'],
    [], 
    { m: true}
  ]
    
  ]
];

describe("mergeWindupsToSuffixed", () => {
  vectors.forEach((vector) => {
    const [title, pair, expected] = vector;

    it(title, () => {
      const [base, against] = pair;

      const suffixedWindup = mergeWindupsToSuffixedWindup(base, against);

      if (isPlayedSuffixedWindup(suffixedWindup)) {
        const [played, suffix, metadata] = suffixedWindup;
        const [ePlayed, eSuffix, eMetadata] = expected;
        expect(played).toEqual(ePlayed);
        expect(suffix).toEqual(eSuffix);
        expect(metadata).toEqual(eMetadata);
      } else {
        const [played, remaining, suffix, metadata] = suffixedWindup;
        const [ePlayed, eRemaining, eSuffix, eMetadata] = expected;
        expect(played).toEqual(ePlayed);
        expect(remaining).toEqual(eRemaining);
        expect(suffix).toEqual(eSuffix);
        expect(metadata).toEqual(eMetadata);
      }
    });
  });
});
