import {
  mergeWindupsToSuffixedWindup,
  SuffixedWindup,
  PlayedSuffixedWindup,
} from "../SuffixedWindup";
import { windupFromString, fastForward } from "../Windup";
import { deepPair, suffixedPrevBy } from "./util";

type TestElement<E, M> = SuffixedWindup<E, M> | PlayedSuffixedWindup<E, M>;

type Expected<E, M> = SuffixedWindup<E, M>;

type Vector<E, M> = [string, TestElement<E, M>, number, Expected<E, M>];

const finishedSentence = fastForward(
  windupFromString("windups can't intelligently diff contents", {
    m: true,
  })
);
const unfinishedSentence = windupFromString(
  "windups can definitely diff their contents",
  {
    m: true,
  }
);

const finishedSuffixed = mergeWindupsToSuffixedWindup(
  finishedSentence,
  unfinishedSentence
);
const unfinishedSuffixed = mergeWindupsToSuffixedWindup(
  unfinishedSentence,
  finishedSentence
);

const deepSuffixed = mergeWindupsToSuffixedWindup(
  deepPair.base,
  deepPair.against
);

const vectors: Vector<string, { m: boolean }>[] = [
  [
    "Played",
    finishedSuffixed,
    1,
    [
      "windups can't intelligently dif".split(""),
      "f".split(""),
      " contents".split(""),
      { m: true },
    ],
  ],

  [
    "Played",
    finishedSuffixed,
    3,
    [
      "windups can't intelligently d".split(""),
      "iff".split(""),
      " contents".split(""),
      { m: true },
    ],
  ],

  [
    "Unplayed",
    unfinishedSuffixed,
    1,
    [
      "windups ca".split(""),
      "n definitely diff their".split(""),
      " contents".split(""),
      { m: true },
    ],
  ],

  [
    "Unplayed",
    unfinishedSuffixed,
    3,
    [
      "windups ".split(""),
      "can definitely diff their".split(""),
      " contents".split(""),
      { m: true },
    ],
  ],
  [
    "Deep",
    deepSuffixed,
    5,
    [
      ["H", "i", "!", " "],
      [
        [
          [],
          ["W", "h", "a", "t", " ", "a", " ", "n", "i", "c", "e"],
          [" ", "r", "o", "o", "m", "y"],
          {
            m: false,
          },
        ],
      ],
      [" ", "p", "l", "a", "c", "e"],
      {
        m: true,
      },
    ],
  ],
];

describe("prev", () => {
  vectors.forEach((vector) => {
    const [title, el, prevBy, expected] = vector;

    it(title, () => {
      const [played, remaining, suffix, metadata] = suffixedPrevBy(el, prevBy);

      const [ePlayed, eRemaining, eSuffix, eMetadata] = expected;

      expect(played).toEqual(ePlayed);
      expect(remaining).toEqual(eRemaining);
      expect(suffix).toEqual(eSuffix);
      expect(metadata).toEqual(eMetadata);
    });
  });
});
