import {
  PlayedSuffixedWindup,
  SuffixedWindup,
  prev as suffixedPrev,
} from "../SuffixedWindup";
import {
  fastForward,
  isPlayedWindup,
  newWindup,
  next,
  PlayedWindup,
  Windup,
  windupFromString,
} from "../Windup";

// Vanilla windups

export const newStrWindup = () => windupFromString("Hello!", { m: true });
export const inProgressStrWindup = () => nextBy(newStrWindup(), 3);

export const playedStrWindup = () => fastForward(newStrWindup());

export const newDeepWindup = (): Windup<string, {}> =>
  newWindup<string, {}>(
    [
      [[], "Hi!".split(""), {}],
      [[], " Bye!".split(""), {}],
    ],
    {}
  );
export const inProgressDeepWindup = () => nextBy(newDeepWindup(), 3);
export const playedDeepWindup = () => fastForward(newDeepWindup());

// Base / Against windups

export const strPair = {
  base: windupFromString("Hi! I am Sam.", { m: true }),
  against: windupFromString("Hi. You are Pam.", { m: true }),
};

export const strPairInProgress = {
  base: nextBy(strPair.base, 3),
  against: nextBy(strPair.against, 3),
};

export const strPairPlayed = {
  base: fastForward(strPair.base),
  against: fastForward(strPair.against),
};

export const deepPair = {
  base: newWindup<string, { m: boolean }>(
    [
      ..."Hi! ".split(""),
      [[], "What a nice roomy".split(""), { m: false }],
      ..." place".split(""),
    ],
    { m: true }
  ),
  against: newWindup<string, { m: boolean }>(
    [
      ..."Hi! ".split(""),
      [[], "What is this roomy".split(""), { m: false }],
      ..." place".split(""),
    ],
    { m: true }
  ),
};

export function nextBy<E, M>(windup: Windup<E, M>, times: number) {
  let res: Windup<E, M> | PlayedWindup<E, M> = windup;

  for (let i = 0; i < times; i++) {
    if (isPlayedWindup(res)) {
      return res;
    }

    res = next(res);
  }

  return res;
}

export function suffixedPrevBy<E, M>(
  windup: SuffixedWindup<E, M> | PlayedSuffixedWindup<E, M>,
  times: number
) {
  let res: SuffixedWindup<E, M> | PlayedSuffixedWindup<E, M> = windup;

  for (let i = 0; i < times; i++) {
    res = suffixedPrev(res);
  }

  return res;
}

export function log(name: string, data: any) {
  console.log(name, JSON.stringify(data, null, 2));
}
